/*
    ***********************************************************************************************
    Sphere rendering code (c) Jarmo 2008-2012
    http://z0b.kapsi.fi
    ***********************************************************************************************

    License
    -------

    This software is provided 'as-is', without any express or implied warranty. In no event
    will the authors be held liable for any damages arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose, including
    commercial applications, and to alter it and redistribute it freely, subject to the
    following restrictions:

        1. The origin of this software must not be misrepresented; you must not
           claim that you wrote the original software. If you use this software
           in a product, an acknowledgment in the product documentation would be
           appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must not be
           misrepresented as being the original software.

        3. This notice may not be removed or altered from any source
           distribution.


    Notes
    -----

    ** SDL IS NOT REQUIRED!! ** It is only used to create a window, read the keyboard and display
    the output! This code is platform-neutral, it has been compiled and tested with GCC 4.5.x,
    under Arch Linux and Windows XP/7.

    All functions clip the sphere fully before rendering. No filtering is done, but should
    not be too hard to be implemented (even a bilinear filtering could dramatically improve
    the output quality).

    The light and sphere rotations are separate (the light can be rotated independently from
    sphere). If you want the light to rotate with the sphere, the best way is to replace lat/lon
    with your custom 3D light vector that you rotate with the rotation matrix.

    Because the functions are almost identical to each other, common parts are commented and
    explained only once. For example, the lit texture function does not explain how the texture
    coordinates are calculated, check the unlit textured sphere function for explanation.

    These functions ooze optimization possibilities. The code is not particularly fast, it is
    only intended to demonstrate the effect. If you seriously want to use any of these functions,
    you HAVE to optimize the shit out of them.

    Thank you, namaste and good luck.


    Common parameters
    -----------------

      'destination': destination bitmap (32-bit)
      'texture': source texture map (24-bit)
      'xp' and 'yp': origin
      'radius': radius, in pixels
      'longitude' and 'latitude': light position, both in 0...2PI range
      'rotation': 4x4 rotation matrix (3x3 works too)
      'red', 'green' and 'blue': light color, 0...255

    ***********************************************************************************************
*/

#include <cstdio>
#include <cmath>
#include <vector>
#include <stdint.h>

using namespace std;

// simple bitmap structure, used to contain output and texture data
struct bitmap_t {
    vector<uint8_t> data;
    int32_t w, h;

    bitmap_t() : w(0), h(0) { }
    ~bitmap_t() { }
};

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// SPHERE RENDERING

// draws a lit sphere, no texturing
void litSphere(bitmap_t *destination,
               const int32_t xp, const int32_t yp,
               const int32_t radius,
               const float longitude, const float latitude,
               const uint8_t red, const uint8_t green, const uint32_t blue)
{
    if (!destination || radius < 1)
        return;

    if (red == 0U && green == 0U && blue == 0U)
        return;

    // reject completely invisible cases
    if (xp + radius <= 0)
        return;

    if (yp + radius <= 0)
        return;

    if (xp - radius >= destination->w - 1)
        return;

    if (yp - radius >= destination->h - 1)
        return;

    // figure out the visible part
    const uint32_t x1 = max(0, xp - radius),
                   x2 = min(destination->w, xp + radius),
                   y1 = max(0, yp - radius),
                   y2 = min(destination->h, yp + radius);

    const int32_t rad2 = radius * radius;

    // compute light's position in 3D space
    const float lx = sin(longitude) * cos(latitude),
                ly = sin(latitude),
                lz = cos(longitude) * cos(latitude);

    // convert color to 0...1 range
    const float lr = (float)red / 255.0f,
                lg = (float)green / 255.0f,
                lb = (float)blue / 255.0f;

    // for every pixel in the area...
    for (uint32_t y = y1; y < y2; y++) {
        for (uint32_t x = x1; x < x2; x++) {
            const int32_t sx = (signed)x - xp,
                          sy = (signed)y - yp;

            // ... is it inside the sphere?
            if ((sy * sy + sx * sx) < rad2) {
                // this pixel is inside, get the Z coordinate
                const float z = sqrt(rad2 - (sx * sx) - (sy * sy));

                // Compute the dot product between it and the light vector.
                // The X, Y and Z coordinates are converted into 0...1 range
                // before multiplications.
                const float light = ((float)sx / (float)radius) * lx +
                                    ((float)sy / (float)radius) * ly +
                                            (z / (float)radius) * lz;

                // shade only the front side
                if (light >= 0.0f) {
                    uint8_t *dst = &destination->data[(y * destination->w + x) * 4];
                    const float lv = light * 255.0f;

                    // multiply light color with light intensity, round and save
                    *dst++ = (uint8_t)(lv * lr);
                    *dst++ = (uint8_t)(lv * lg);
                    *dst++ = (uint8_t)(lv * lb);
                }
            }
        }
    }
}

// draws an unlit textured sphere
void texturedSphere(bitmap_t *destination, const bitmap_t *texture,
                    const int32_t xp, const int32_t yp,
                    const int32_t radius,
                    const float *rotation)
{
    if (!destination || !texture || radius < 1 || !rotation)
        return;

    // reject completely invisible cases
    if (xp + radius <= 0)
        return;

    if (yp + radius <= 0)
        return;

    if (xp - radius >= destination->w - 1)
        return;

    if (yp - radius >= destination->h - 1)
        return;

    const uint32_t x1 = max(0, xp - radius),
                   x2 = min(destination->w, xp + radius),
                   y1 = max(0, yp - radius),
                   y2 = min(destination->h, yp + radius);

    const int32_t rad2 = radius * radius;

    const float fRad = (float)radius;

    for (uint32_t y = y1; y < y2; y++) {
        for (uint32_t x = x1; x < x2; x++) {
            const int32_t sx = (signed)x - xp,
                          sy = (signed)y - yp;

            if ((sy * sy + sx * sx) < rad2) {
                // get the Z coordinate
                const float z = sqrt(rad2 - sx * sx - sy * sy);

                // rotate the 3D point and scale to 0...1 range (for trig functions)
                const float
                    rx = (sx * rotation[0] + sy * rotation[1] + z * rotation[2] + rotation[3]) / fRad,
                    ry = (sx * rotation[4] + sy * rotation[5] + z * rotation[6] + rotation[7]) / fRad,
                    rz = (sx * rotation[8] + sy * rotation[9] + z * rotation[10] + rotation[11]) / fRad;

                // Get X coordinate for texture mapping. atan2() returns a value
                // between -PI and PI, so we scale that to 0...2PI, then divide
                // it by two and scale to 0...(textureW - 1).
                const int32_t tx = (int32_t)((atan2(rx, rz) + M_PI) / (2.0f * M_PI) * (float)(texture->w - 1));

                // Then get the texture Y coordinate. asin() returns a value
                // between -PI/2 and PI/2, so we add PI/2, divide by PI and scale
                // to 0...(textureH - 1).
                const int32_t ty = (int32_t)((asin(ry) + (M_PI / 2.0f)) / M_PI * (float)(texture->h - 1));

                // copy pixel
                const uint8_t *src = &texture->data[(ty * texture->w + tx) * 3];
                uint8_t *dst = &destination->data[(y * destination->w + x) * 4];

                *dst++ = *src++;
                *dst++ = *src++;
                *dst++ = *src++;
            }
        }
    }
}

// draws a lit and textured sphere
void litTexturedSphere(bitmap_t *destination, const bitmap_t *texture,
                       const int32_t xp, const int32_t yp,
                       const int32_t radius,
                       const float *rotation,
                       const float longitude, const float latitude,
                       const uint8_t red, const uint8_t green, const uint32_t blue)
{
    if (!destination || !texture || radius < 1 || !rotation)
        return;

    if (red == 0U && green == 0U && blue == 0U)
        return;

    // reject completely invisible cases
    if (xp + radius <= 0)
        return;

    if (yp + radius <= 0)
        return;

    if (xp - radius >= destination->w - 1)
        return;

    if (yp - radius >= destination->h - 1)
        return;

    const uint32_t x1 = max(0, xp - radius),
                   x2 = min(destination->w, xp + radius),
                   y1 = max(0, yp - radius),
                   y2 = min(destination->h, yp + radius);

    const int32_t rad2 = radius * radius;

    // compute light's position in 3D space
    const float lx = sin(longitude) * cos(latitude),
                ly = sin(latitude),
                lz = cos(longitude) * cos(latitude);

    // convert color to 0...1 range
    const float lr = (float)red / 255.0f,
                lg = (float)green / 255.0f,
                lb = (float)blue / 255.0f;

    const float fRad = (float)radius;

    for (uint32_t y = y1; y < y2; y++) {
        for (uint32_t x = x1; x < x2; x++) {
            const int32_t sx = (signed)x - xp,
                          sy = (signed)y - yp;

            if ((sy * sy + sx * sx) < rad2) {
                const float z = sqrt(rad2 - sx * sx - sy * sy),
                            light = ((float)sx / fRad) * lx +
                                    ((float)sy / fRad) * ly +
                                            (z / fRad) * lz;

                if (light >= 0.0f) {
                    const float
                        rx = (sx * rotation[0] + sy * rotation[1] + z * rotation[2] + rotation[3]) / fRad,
                        ry = (sx * rotation[4] + sy * rotation[5] + z * rotation[6] + rotation[7]) / fRad,
                        rz = (sx * rotation[8] + sy * rotation[9] + z * rotation[10] + rotation[11]) / fRad;

                    const int32_t tx = (int32_t)((atan2(rx, rz) + M_PI) / (2.0f * M_PI) *
                                       (float)(texture->w - 1)),
                                  ty = (int32_t)((asin(ry) + (M_PI / 2.0f)) / M_PI *
                                       (float)(texture->h - 1));

                    // get pixel, colorize, shade and write it into destination bitmap
                    // (convert each component from 0...255 to 0..1, saturate, adjust
                    // brightness, then scale back to 0...255)
                    const uint8_t *src = &texture->data[(ty * texture->w + tx) * 3];
                    uint8_t *dst = &destination->data[(y * destination->w + x) * 4];

                    *dst++ = (uint8_t)(((*src++ / 255.0f) * lr * light) * 255.0f);
                    *dst++ = (uint8_t)(((*src++ / 255.0f) * lg * light) * 255.0f);
                    *dst++ = (uint8_t)(((*src++ / 255.0f) * lb * light) * 255.0f);
                }
            }
        }
    }
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// SIMPLE DEMO

// Only to demonstrate the functions!

#include <SDL/SDL.h>
#include <cstdarg>

static const uint32_t screenW = 512U, screenH = 384U;

// do not use this, SDL does not have built-in text output so I scraped one together
static void cprintf(uint8_t *dest, const uint32_t destW, const uint32_t destH,
                    const uint32_t x, const uint32_t y, const char *fmt, ...)
{
    if (!dest || !fmt)
        return;

    // 8x16 font data, only 96 characters (partial 7-bit ASCII)
    static const uint8_t fontData[96 * 16] =
    {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x18, 0x3C, 0x3C, 0x3C, 0x18, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x66, 0x66, 0x66, 0x24, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x6C, 0x6C, 0xFE, 0x6C, 0x6C, 0x6C, 0xFE, 0x6C, 0x6C, 0x00, 0x00, 0x00, 0x00,
        0x18, 0x18, 0x7C, 0xC6, 0xC2, 0xC0, 0x7C, 0x06, 0x06, 0x86, 0xC6, 0x7C, 0x18, 0x18, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0xC2, 0xC6, 0x0C, 0x18, 0x30, 0x60, 0xC6, 0x86, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x38, 0x6C, 0x6C, 0x38, 0x76, 0xDC, 0xCC, 0xCC, 0xCC, 0x76, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x30, 0x30, 0x30, 0x60, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x0C, 0x18, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x18, 0x0C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x30, 0x18, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x18, 0x30, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x66, 0x3C, 0xFF, 0x3C, 0x66, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x7E, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x18, 0x30, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFE, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x02, 0x06, 0x0C, 0x18, 0x30, 0x60, 0xC0, 0x80, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x38, 0x6C, 0xC6, 0xC6, 0xD6, 0xD6, 0xC6, 0xC6, 0x6C, 0x38, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x18, 0x38, 0x78, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x7E, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x7C, 0xC6, 0x06, 0x0C, 0x18, 0x30, 0x60, 0xC0, 0xC6, 0xFE, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x7C, 0xC6, 0x06, 0x06, 0x3C, 0x06, 0x06, 0x06, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x0C, 0x1C, 0x3C, 0x6C, 0xCC, 0xFE, 0x0C, 0x0C, 0x0C, 0x1E, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xFE, 0xC0, 0xC0, 0xC0, 0xFC, 0x06, 0x06, 0x06, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x38, 0x60, 0xC0, 0xC0, 0xFC, 0xC6, 0xC6, 0xC6, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xFE, 0xC6, 0x06, 0x06, 0x0C, 0x18, 0x30, 0x30, 0x30, 0x30, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x7C, 0xC6, 0xC6, 0xC6, 0x7C, 0xC6, 0xC6, 0xC6, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x7C, 0xC6, 0xC6, 0xC6, 0x7E, 0x06, 0x06, 0x06, 0x0C, 0x78, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x18, 0x18, 0x30, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x06, 0x0C, 0x18, 0x30, 0x60, 0x30, 0x18, 0x0C, 0x06, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x7E, 0x00, 0x00, 0x7E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x60, 0x30, 0x18, 0x0C, 0x06, 0x0C, 0x18, 0x30, 0x60, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x7C, 0xC6, 0xC6, 0x0C, 0x18, 0x18, 0x18, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x7C, 0xC6, 0xC6, 0xDE, 0xDE, 0xDE, 0xDC, 0xC0, 0x7C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x10, 0x38, 0x6C, 0xC6, 0xC6, 0xFE, 0xC6, 0xC6, 0xC6, 0xC6, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xFC, 0x66, 0x66, 0x66, 0x7C, 0x66, 0x66, 0x66, 0x66, 0xFC, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x3C, 0x66, 0xC2, 0xC0, 0xC0, 0xC0, 0xC0, 0xC2, 0x66, 0x3C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xF8, 0x6C, 0x66, 0x66, 0x66, 0x66, 0x66, 0x66, 0x6C, 0xF8, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xFE, 0x66, 0x62, 0x68, 0x78, 0x68, 0x60, 0x62, 0x66, 0xFE, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xFE, 0x66, 0x62, 0x68, 0x78, 0x68, 0x60, 0x60, 0x60, 0xF0, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x3C, 0x66, 0xC2, 0xC0, 0xC0, 0xDE, 0xC6, 0xC6, 0x66, 0x3A, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xC6, 0xC6, 0xC6, 0xC6, 0xFE, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x3C, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x1E, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0xCC, 0xCC, 0xCC, 0x78, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xE6, 0x66, 0x66, 0x6C, 0x78, 0x78, 0x6C, 0x66, 0x66, 0xE6, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xF0, 0x60, 0x60, 0x60, 0x60, 0x60, 0x60, 0x62, 0x66, 0xFE, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xC6, 0xEE, 0xFE, 0xFE, 0xD6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xC6, 0xE6, 0xF6, 0xFE, 0xDE, 0xCE, 0xC6, 0xC6, 0xC6, 0xC6, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x7C, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xFC, 0x66, 0x66, 0x66, 0x7C, 0x60, 0x60, 0x60, 0x60, 0xF0, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x7C, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xD6, 0xDE, 0x7C, 0x0C, 0x0E, 0x00, 0x00,
        0x00, 0x00, 0xFC, 0x66, 0x66, 0x66, 0x7C, 0x6C, 0x66, 0x66, 0x66, 0xE6, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x7C, 0xC6, 0xC6, 0x60, 0x38, 0x0C, 0x06, 0xC6, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x7E, 0x7E, 0x5A, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x6C, 0x38, 0x10, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xC6, 0xC6, 0xC6, 0xC6, 0xD6, 0xD6, 0xD6, 0xFE, 0xEE, 0x6C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xC6, 0xC6, 0x6C, 0x7C, 0x38, 0x38, 0x7C, 0x6C, 0xC6, 0xC6, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x66, 0x66, 0x66, 0x66, 0x3C, 0x18, 0x18, 0x18, 0x18, 0x3C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xFE, 0xC6, 0x86, 0x0C, 0x18, 0x30, 0x60, 0xC2, 0xC6, 0xFE, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x3C, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x3C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x80, 0xC0, 0xE0, 0x70, 0x38, 0x1C, 0x0E, 0x06, 0x02, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x3C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x3C, 0x00, 0x00, 0x00, 0x00,
        0x10, 0x38, 0x6C, 0xC6, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00,
        0x30, 0x30, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x78, 0x0C, 0x7C, 0xCC, 0xCC, 0xCC, 0x76, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0xE0, 0x60, 0x60, 0x78, 0x6C, 0x66, 0x66, 0x66, 0x66, 0x7C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x7C, 0xC6, 0xC0, 0xC0, 0xC0, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x1C, 0x0C, 0x0C, 0x3C, 0x6C, 0xCC, 0xCC, 0xCC, 0xCC, 0x76, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x7C, 0xC6, 0xFE, 0xC0, 0xC0, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x38, 0x6C, 0x64, 0x60, 0xF0, 0x60, 0x60, 0x60, 0x60, 0xF0, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x76, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0x7C, 0x0C, 0xCC, 0x78, 0x00,
        0x00, 0x00, 0xE0, 0x60, 0x60, 0x6C, 0x76, 0x66, 0x66, 0x66, 0x66, 0xE6, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x18, 0x18, 0x00, 0x38, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x06, 0x06, 0x00, 0x0E, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x66, 0x66, 0x3C, 0x00,
        0x00, 0x00, 0xE0, 0x60, 0x60, 0x66, 0x6C, 0x78, 0x78, 0x6C, 0x66, 0xE6, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x38, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0xEC, 0xFE, 0xD6, 0xD6, 0xD6, 0xD6, 0xC6, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0xDC, 0x66, 0x66, 0x66, 0x66, 0x66, 0x66, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x7C, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0xDC, 0x66, 0x66, 0x66, 0x66, 0x66, 0x7C, 0x60, 0x60, 0xF0, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x76, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0x7C, 0x0C, 0x0C, 0x1E, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0xDC, 0x76, 0x66, 0x60, 0x60, 0x60, 0xF0, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x7C, 0xC6, 0x60, 0x38, 0x0C, 0xC6, 0x7C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x10, 0x30, 0x30, 0xFC, 0x30, 0x30, 0x30, 0x30, 0x36, 0x1C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0x76, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x66, 0x66, 0x66, 0x66, 0x66, 0x3C, 0x18, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0xC6, 0xC6, 0xD6, 0xD6, 0xD6, 0xFE, 0x6C, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0xC6, 0x6C, 0x38, 0x38, 0x38, 0x6C, 0xC6, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x7E, 0x06, 0x0C, 0xF8, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0xFE, 0xCC, 0x18, 0x30, 0x60, 0xC6, 0xFE, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x0E, 0x18, 0x18, 0x18, 0x70, 0x18, 0x18, 0x18, 0x18, 0x0E, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x70, 0x18, 0x18, 0x18, 0x0E, 0x18, 0x18, 0x18, 0x18, 0x70, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x76, 0xDC, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x10, 0x38, 0x6C, 0xC6, 0xC6, 0xC6, 0xFE, 0x00, 0x00, 0x00, 0x00, 0x00,
    };

#define COLOR(r, g, b)      ((r << 16) | (g << 8) | (b))

    // colors
    static const uint32_t colors[16] =
    {
        COLOR(0, 0, 0),         // 0
        COLOR(0, 0, 168),       // 1
        COLOR(0, 168, 0),       // 2
        COLOR(0, 168, 168),     // 3
        COLOR(168, 0, 0),       // 4
        COLOR(168, 0, 168),     // 5
        COLOR(168, 168, 0),     // 6
        COLOR(168, 168, 168),   // 7
        COLOR(84, 84, 84),      // 8
        COLOR(84, 84, 252),     // 9
        COLOR(84, 252, 84),     // 10
        COLOR(84, 252, 252),    // 11
        COLOR(252, 84, 84),     // 12
        COLOR(252, 84, 252),    // 13
        COLOR(252, 252, 84),    // 14
        COLOR(252, 252, 252)    // 15
    };

#undef COLOR

    const uint32_t rowDelta = destW - 8U;
    uint32_t fore = colors[7],
             xp = x, yp = y;

    static char buf[1024];
    va_list ap;

    va_start(ap, fmt);

    if (vsnprintf(buf, 1023, fmt, ap) == -1)
        return;

    for (const char *p = buf; *p; p++) {
        if (*p == '\n') {
            xp = x;
            yp += 16;
            continue;
        } else if (*p == '$' && *(p + 1)) {
            const char c = *(p + 1);

            if (c >= '0' && c <= '9')
                fore = colors[c - '0'];
            else if (c >= 'A' && c <= 'F')
                fore = colors[c - 'A' + 10];
            else if (c >= 'a' && c <= 'f')
                fore = colors[c - 'a' + 10];
            else fore = 7;

            p++;
            continue;
        } else if (*p < 32 || *p > 128)
            continue;

        if (yp > destH)
            break;

        if (xp < destW) {
            const uint8_t *font = &fontData[(*p - 32) << 4];

            uint32_t *dst = (uint32_t *)(dest + ((yp * destW + xp) * 4)),
                     y = 16U;

            while (y--) {
                if (*font & 0x80) *dst = fore; dst++;
                if (*font & 0x40) *dst = fore; dst++;
                if (*font & 0x20) *dst = fore; dst++;
                if (*font & 0x10) *dst = fore; dst++;
                if (*font & 0x08) *dst = fore; dst++;
                if (*font & 0x04) *dst = fore; dst++;
                if (*font & 0x02) *dst = fore; dst++;
                if (*font & 0x01) *dst = fore; dst++;

                dst += rowDelta;
                ++font;
            }
        }

        xp += 8;
    }
}

// constructs a 4x4 rotation matrix
static void rotationMatrix(const float x, const float y, const float z, float *out)
{
    if (!out)
        return;

    const float degToRad = 0.017453292f,
                rx =x * degToRad, ry = y * degToRad, rz = z * degToRad,
                sx = sin(rx), cx = cos(rx),
                sy = sin(ry), cy = cos(ry),
                sz = sin(rz), cz = cos(rz);

    out[0] = cy * cz;
    out[1] = cy * sz;
    out[2] = -sy;
    out[3] = 0.0;
    out[4] = (sx *sy *cz) + (cx * -sz);
    out[5] = (sx *sy *sz) + (cx * cz);
    out[6] = sx * cy;
    out[7] = 0.0;
    out[8] = (cx * sy * cz) + (-sx * -sz);
    out[9] = (cx * sy * sz) + (-sx * cz);
    out[10] = cx * cy;
    out[11] = 0.0;
    out[12] = 0.0;
    out[13] = 0.0;
    out[14] = 0.0;
    out[15] = 1.0;
}

// matrix multiplication
static void matrixMul(const float *m1, const float *m2, float *out)
{
    if (!m1 || !m2 || !out)
        return;

    for (int i = 0; i < 4; i++) {
        out[i + 0] = m1[i + 0] * m2[0] + m1[i + 4] * m2[1] + m1[i + 8] * m2[2] + m1[i + 12] * m2[3];
        out[i + 4] = m1[i + 0] * m2[4] + m1[i + 4] * m2[5] + m1[i + 8] * m2[6] + m1[i + 12] * m2[7];
        out[i + 8] = m1[i + 0] * m2[8] + m1[i + 4] * m2[9] + m1[i + 8] * m2[10] + m1[i + 12] * m2[11];
        out[i + 12] = m1[i + 0] * m2[12] + m1[i + 4] * m2[13] + m1[i + 8] * m2[14] + m1[i + 12] * m2[15];
    }
}

int main(int argc, char *argv[])
{
    // init SDL
    if (SDL_Init(SDL_INIT_VIDEO)) {
        fprintf(stderr, "SDL init failed: %s\n", SDL_GetError());
        return 1;
    }

    SDL_Surface *screen = SDL_SetVideoMode(screenW, screenH, 32, SDL_SWSURFACE);

    if (!screen) {
        fprintf(stderr, "Unable to set 400x300 video: %s\n", SDL_GetError());
        SDL_Quit();
        return 1;
    }

    SDL_WM_SetCaption("Sphere Rendering (c) Jarmo", NULL);
    SDL_EnableKeyRepeat(200, 10);

    // load the texture
    bitmap_t tex;

    tex.w = 512;
    tex.h = 256;
    tex.data.resize(tex.w * tex.h * 3);

    FILE *f = fopen("texture.raw", "rb");

    if (!f) {
        fprintf(stderr, "Can't load the texture\n");
        SDL_Quit();
        return 1;
    }

    fread(&tex.data[0], 1, tex.data.size(), f);
    fclose(f);

    // destination bitmap
    bitmap_t buffer;

    buffer.w = 256;
    buffer.h = 256;
    buffer.data.resize(buffer.w * buffer.h * 4);

    static const char *renderMode[] = { "Lit untextured", "Unlit textured", "Lit textured" };

    bool done = false,
         drag = false;

    float xr = 0.0f, yr = 0.0f,
          longitude = 0.0f, latitude = 0.0f;
    int32_t radius = 128,
            posX = buffer.w / 2, posY = buffer.h / 2,
            style = 1;

    // main loop
    while (!done) {
        // create rotation matrix (first X then Y, it looks better that way)
        float xRot[16], yRot[16], rot[16];

        rotationMatrix(xr, 0.0f, 0.0f, xRot);
        rotationMatrix(0.0f, yr, 0.0f, yRot);
        matrixMul(xRot, yRot, rot);

        // render
        memset(&buffer.data[0], 0, buffer.data.size());

        switch (style) {
            case 0:
                litSphere(&buffer, posX, posY, radius, latitude, longitude, 255, 255, 255);
                break;

            case 1:
                texturedSphere(&buffer, &tex, posX, posY, radius, rot);
                break;

            case 2:
                litTexturedSphere(&buffer, &tex, posX, posY, radius, rot, latitude, longitude, 255, 255, 255);
                break;
        }

        if (SDL_MUSTLOCK(screen))
            if (!SDL_LockSurface(screen)) {
                fprintf(stderr, "SDL_LockSurface() failed: %s\n", SDL_GetError());
                break;
            }

        uint8_t *pixels = (uint8_t *)screen->pixels;

        memset(pixels, 0, screenW * screenH * 4);

        // copy sphere to screen
        const uint32_t top = screenH / 2 - buffer.h / 2,
                       left = screenW / 2 - buffer.w / 2;

        for (int32_t y = 0; y < buffer.h; y++) {
            uint8_t *dest = pixels + ((top + y) * screenW + left) * 4;
            const uint8_t *src = &buffer.data[y * buffer.w * 4];

            for (int32_t x = 0; x < buffer.w; x++) {
                *dest++ = src[2];
                *dest++ = src[1];
                *dest++ = src[0];
                *dest++ = src[3];
                src += 4;
            }
        }

        // draw a rectangle (SDL WHY U NO PRIMITIVES BUILT-IN?)
        memset(pixels + ((top - 2) * screenW + left - 1) * 4, 128, (buffer.w + 2) * 4);
        memset(pixels + ((top + buffer.h + 1) * screenW + left - 1) * 4, 128, (buffer.w + 2) * 4);

        for (uint32_t y = top - 2; y < top + buffer.h + 2; y++) {
            uint8_t *dest;

            dest = pixels + (y * screenW + left - 2) * 4;
            *dest++ = 128;
            *dest++ = 128;
            *dest++ = 128;

            dest = pixels + (y * screenW + left + buffer.w + 1) * 4;
            *dest++ = 128;
            *dest++ = 128;
            *dest++ = 128;
        }

        cprintf(pixels, screenW, screenH, 0, 0,
                "$FRotation: $AX=$C%.1f  $AY=$C%.1f  $FRadius: $C%d$7/255\n$FLight: $Alat=$C%.2f $Alon=$C%.2f",
                xr, yr, radius, latitude, longitude);

        cprintf(pixels, screenW, screenH, 0, screenH - 32,
                "$EArrows$7=rotate sphere, $EWASD$7=rotate light, $Ekeypad +/-$7=radius\n$ETAB$7=rendering style [$B%s$7]",
                renderMode[style]);

        if (SDL_MUSTLOCK(screen))
            SDL_UnlockSurface(screen);

        SDL_UpdateRect(screen, 0, 0, screenW, screenH);

        // handle events, if any
        SDL_Event event;

        while (SDL_PollEvent(&event)) {
            switch (event.type) {
                case SDL_KEYDOWN:
                {
                    switch (event.key.keysym.sym) {
                        case SDLK_ESCAPE:
                            done = true;
                            break;

                        // rendering style
                        case SDLK_TAB:
                            style = (style + 1) % 3;
                            break;

                        // change radius
                        case SDLK_KP_PLUS:
                            radius = min(255, radius + 1);
                            break;

                        case SDLK_KP_MINUS:
                            radius = max(0, radius - 1);
                            break;

                        // rotate sphere
                        case SDLK_LEFT:
                            yr -= 1.0f;
                            break;

                        case SDLK_RIGHT:
                            yr += 1.0f;
                            break;

                        case SDLK_UP:
                            xr -= 1.0f;
                            break;

                        case SDLK_DOWN:
                            xr += 1.0f;
                            break;

                        // change light position
                        case SDLK_w:
                            longitude -= 0.01f;
                            break;

                        case SDLK_s:
                            longitude += 0.01f;
                            break;

                        case SDLK_a:
                            latitude -= 0.01f;
                            break;

                        case SDLK_d:
                            latitude += 0.01f;
                            break;

                        default:
                            break;
                    }

                    break;
                }

                // begin drag
                case SDL_MOUSEBUTTONDOWN:
                    if (event.button.button == SDL_BUTTON_LEFT)
                        drag = true;

                    break;

                // end drag
                case SDL_MOUSEBUTTONUP:
                    if (event.button.button == SDL_BUTTON_LEFT)
                        drag = false;

                    break;

                // drag
                case SDL_MOUSEMOTION:
                    if (drag) {
                        posX += event.motion.xrel;
                        posY += event.motion.yrel;
                    }

                    break;

                case SDL_QUIT:
                    done = true;
                    break;
            }
        }
    }

    SDL_Quit();
    return 0;
}
